@@ -1,5 +1,6 @@
 #define _POSIX_C_SOURCE 200809L
 #include <ctype.h>
+#include <math.h>
 #include <stdbool.h>
 #include <string.h>
 #include <strings.h>
@@ -112,38 +113,41 @@ static void container_move_to_container_from_direction(
 		struct sway_container *container, struct sway_container *destination,
 		enum wlr_direction move_dir) {
 	if (destination->view) {
-		if (destination->parent == container->parent &&
-				destination->workspace == container->workspace) {
+		if (destination->pending.parent == container->pending.parent &&
+				destination->pending.workspace == container->pending.workspace) {
 			sway_log(SWAY_DEBUG, "Swapping siblings");
 			list_t *siblings = container_get_siblings(container);
 			int container_index = list_find(siblings, container);
 			int destination_index = list_find(siblings, destination);
 			list_swap(siblings, container_index, destination_index);
+			container_update_representation(container);
 		} else {
 			sway_log(SWAY_DEBUG, "Promoting to sibling of cousin");
 			int offset =
 				move_dir == WLR_DIRECTION_LEFT || move_dir == WLR_DIRECTION_UP;
 			int index = container_sibling_index(destination) + offset;
-			if (destination->parent) {
-				container_insert_child(destination->parent, container, index);
+			if (destination->pending.parent) {
+				container_insert_child(destination->pending.parent, container, index);
 			} else {
-				workspace_insert_tiling(destination->workspace,
+				workspace_insert_tiling(destination->pending.workspace,
 						container, index);
 			}
-			container->width = container->height = 0;
+			container->pending.width = container->pending.height = 0;
 			container->width_fraction = container->height_fraction = 0;
+			workspace_squash(destination->pending.workspace);
 		}
 		return;
 	}
 
-	if (is_parallel(destination->layout, move_dir)) {
+	if (is_parallel(destination->pending.layout, move_dir)) {
 		sway_log(SWAY_DEBUG, "Reparenting container (parallel)");
 		int index =
 			move_dir == WLR_DIRECTION_RIGHT || move_dir == WLR_DIRECTION_DOWN ?
-			0 : destination->children->length;
+			0 : destination->pending.children->length;
 		container_insert_child(destination, container, index);
-		container->width = container->height = 0;
+		container->pending.width = container->pending.height = 0;
 		container->width_fraction = container->height_fraction = 0;
+		workspace_squash(destination->pending.workspace);
 		return;
 	}
 
@@ -164,7 +168,7 @@ static void container_move_to_container_from_direction(
 static void container_move_to_workspace_from_direction(
 		struct sway_container *container, struct sway_workspace *workspace,
 		enum wlr_direction move_dir) {
-	container->width = container->height = 0;
+	container->pending.width = container->pending.height = 0;
 	container->width_fraction = container->height_fraction = 0;
 
 	if (is_parallel(workspace->layout, move_dir)) {
@@ -184,8 +188,8 @@ static void container_move_to_workspace_from_direction(
 		workspace_add_tiling(workspace, container);
 		return;
 	}
-	while (focus_inactive->parent) {
-		focus_inactive = focus_inactive->parent;
+	while (focus_inactive->pending.parent) {
+		focus_inactive = focus_inactive->pending.parent;
 	}
 	container_move_to_container_from_direction(container, focus_inactive,
 			move_dir);
@@ -193,25 +197,25 @@ static void container_move_to_workspace_from_direction(
 
 static void container_move_to_workspace(struct sway_container *container,
 		struct sway_workspace *workspace) {
-	if (container->workspace == workspace) {
+	if (container->pending.workspace == workspace) {
 		return;
 	}
-	struct sway_workspace *old_workspace = container->workspace;
+	struct sway_workspace *old_workspace = container->pending.workspace;
 	if (container_is_floating(container)) {
-		struct sway_output *old_output = container->workspace->output;
+		struct sway_output *old_output = container->pending.workspace->output;
 		container_detach(container);
 		workspace_add_floating(workspace, container);
 		container_handle_fullscreen_reparent(container);
 		// If changing output, center it within the workspace
-		if (old_output != workspace->output && !container->fullscreen_mode) {
+		if (old_output != workspace->output && !container->pending.fullscreen_mode) {
 			container_floating_move_to_center(container);
 		}
 	} else {
 		container_detach(container);
-		if (workspace_is_empty(workspace) && container->children) {
+		if (workspace_is_empty(workspace) && container->pending.children) {
 			workspace_unwrap_children(workspace, container);
 		} else {
-			container->width = container->height = 0;
+			container->pending.width = container->pending.height = 0;
 			container->width_fraction = container->height_fraction = 0;
 			workspace_add_tiling(workspace, container);
 		}
@@ -233,13 +237,13 @@ static void container_move_to_container(struct sway_container *container,
 		return;
 	}
 	if (container_is_floating(container)) {
-		container_move_to_workspace(container, destination->workspace);
+		container_move_to_workspace(container, destination->pending.workspace);
 		return;
 	}
-	struct sway_workspace *old_workspace = container->workspace;
+	struct sway_workspace *old_workspace = container->pending.workspace;
 
 	container_detach(container);
-	container->width = container->height = 0;
+	container->pending.width = container->pending.height = 0;
 	container->width_fraction = container->height_fraction = 0;
 
 	if (destination->view) {
@@ -252,147 +256,154 @@ static void container_move_to_container(struct sway_container *container,
 		ipc_event_window(container, "move");
 	}
 
-	if (destination->workspace) {
-		workspace_focus_fullscreen(destination->workspace);
-		workspace_detect_urgent(destination->workspace);
+	if (destination->pending.workspace) {
+		workspace_focus_fullscreen(destination->pending.workspace);
+		workspace_detect_urgent(destination->pending.workspace);
 	}
 
-	if (old_workspace && old_workspace != destination->workspace) {
+	if (old_workspace && old_workspace != destination->pending.workspace) {
 		workspace_detect_urgent(old_workspace);
 	}
 }
 
-/* Takes one child, sets it aside, wraps the rest of the children in a new
- * container, switches the layout of the workspace, and drops the child back in.
- * In other words, rejigger it. */
-static void workspace_rejigger(struct sway_workspace *ws,
-		struct sway_container *child, enum wlr_direction move_dir) {
-	if (!child->parent && ws->tiling->length == 1) {
-		ws->layout =
-			move_dir == WLR_DIRECTION_LEFT || move_dir == WLR_DIRECTION_RIGHT ?
-			L_HORIZ : L_VERT;
-		workspace_update_representation(ws);
-		return;
+static bool container_move_to_next_output(struct sway_container *container,
+		struct sway_output *output, enum wlr_direction move_dir) {
+	struct sway_output *next_output =
+		output_get_in_direction(output, move_dir);
+	if (next_output) {
+		struct sway_workspace *ws = output_get_active_workspace(next_output);
+		if (!sway_assert(ws, "Expected output to have a workspace")) {
+			return false;
+		}
+		switch (container->pending.fullscreen_mode) {
+		case FULLSCREEN_NONE:
+			container_move_to_workspace_from_direction(container, ws, move_dir);
+			return true;
+		case FULLSCREEN_WORKSPACE:
+			container_move_to_workspace(container, ws);
+			return true;
+		case FULLSCREEN_GLOBAL:
+			return false;
+		}
 	}
-	container_detach(child);
-	workspace_wrap_children(ws);
-
-	int index =
-		move_dir == WLR_DIRECTION_LEFT || move_dir == WLR_DIRECTION_UP ? 0 : 1;
-	workspace_insert_tiling(ws, child, index);
-	ws->layout =
-		move_dir == WLR_DIRECTION_LEFT || move_dir == WLR_DIRECTION_RIGHT ?
-		L_HORIZ : L_VERT;
-	workspace_update_representation(ws);
-	child->width = child->height = 0;
+	return false;
 }
 
 // Returns true if moved
 static bool container_move_in_direction(struct sway_container *container,
 		enum wlr_direction move_dir) {
 	// If moving a fullscreen view, only consider outputs
-	if (container->fullscreen_mode == FULLSCREEN_WORKSPACE) {
-		struct sway_output *new_output =
-			output_get_in_direction(container->workspace->output, move_dir);
-		if (!new_output) {
-			return false;
-		}
-		struct sway_workspace *ws = output_get_active_workspace(new_output);
-		if (!sway_assert(ws, "Expected output to have a workspace")) {
-			return false;
-		}
-		container_move_to_workspace(container, ws);
-		return true;
-	}
-	if (container->fullscreen_mode == FULLSCREEN_GLOBAL) {
+	switch (container->pending.fullscreen_mode) {
+	case FULLSCREEN_NONE:
+		break;
+	case FULLSCREEN_WORKSPACE:
+		return container_move_to_next_output(container,
+				container->pending.workspace->output, move_dir);
+	case FULLSCREEN_GLOBAL:
 		return false;
 	}
 
-	// If container is in a split container by itself, move out of the split
-	if (container->parent) {
-		struct sway_container *new_parent =
-			container_flatten(container->parent);
-		if (new_parent != container->parent) {
-			return true;
-		}
-	}
-
-	// Look for a suitable *container* sibling or parent.
-	// The below loop stops once we hit the workspace because current->parent
-	// is NULL for the topmost containers in a workspace.
-	struct sway_container *current = container;
 	int offs =
 		move_dir == WLR_DIRECTION_LEFT || move_dir == WLR_DIRECTION_UP ? -1 : 1;
+	int index = -1;
+	int	desired = -1;
+	list_t *siblings = NULL;
+	struct sway_container *target = NULL;
 
-	while (current) {
-		list_t *siblings = container_get_siblings(current);
-		if (siblings) {
-			enum sway_container_layout layout = container_parent_layout(current);
-			int index = list_find(siblings, current);
-			int desired = index + offs;
+	// Look for a suitable ancestor of the container to move within
+	struct sway_container *ancestor = NULL;
+	struct sway_container *current = container;
+	bool wrapped = false;
+	while (!ancestor) {
+		// Don't allow containers to move out of their
+		// fullscreen or floating parent
+		if (current->pending.fullscreen_mode || container_is_floating(current)) {
+			return false;
+		}
 
-			// Don't allow containers to move out of their
-			// fullscreen or floating parent
-			if (current->fullscreen_mode || container_is_floating(current)) {
-				return false;
+		enum sway_container_layout parent_layout = container_parent_layout(current);
+		if (!is_parallel(parent_layout, move_dir)) {
+			if (!current->pending.parent) {
+				// No parallel parent, so we reorient the workspace
+				current = workspace_wrap_children(current->pending.workspace);
+				current->pending.workspace->layout =
+					move_dir == WLR_DIRECTION_LEFT ||
+					move_dir == WLR_DIRECTION_RIGHT ?
+					L_HORIZ : L_VERT;
+				container->pending.height = container->pending.width = 0;
+				container->height_fraction = container->width_fraction = 0;
+				workspace_update_representation(current->pending.workspace);
+				wrapped = true;
+			} else {
+				// Keep looking for a parallel parent
+				current = current->pending.parent;
 			}
+			continue;
+		}
 
-			if (is_parallel(layout, move_dir)) {
-				if (desired == -1 || desired == siblings->length) {
-					if (current->parent == container->parent) {
-						current = current->parent;
-						continue;
-					} else {
-						// Reparenting
-						if (current->parent) {
-							container_insert_child(current->parent, container,
-									index + (offs < 0 ? 0 : 1));
-						} else {
-							struct sway_workspace *ws = current->workspace;
-							workspace_insert_tiling(ws,
-								container_split(container,
-									output_get_default_layout(ws->output)),
-								index + (offs < 0 ? 0 : 1));
-						}
-						return true;
-					}
-				} else {
-					// Container can move within its siblings
-					container_move_to_container_from_direction(container,
-							siblings->items[desired], move_dir);
-					return true;
-				}
+		// Only scratchpad hidden containers don't have siblings
+		// so siblings != NULL here
+		siblings = container_get_siblings(current);
+		index = list_find(siblings, current);
+		desired = index + offs;
+		target = desired == -1 || desired == siblings->length ?
+				NULL : siblings->items[desired];
+
+		// If the move is simple we can complete it here early
+		if (current == container) {
+			if (target) {
+				// Container will swap with or descend into its neighbor
+				container_move_to_container_from_direction(container,
+						target, move_dir);
+				return true;
+			} else if (!container->pending.parent) {
+				// Container is at workspace level so we move it to the
+				// next workspace if possible
+				return container_move_to_next_output(container,
+						current->pending.workspace->output, move_dir);
+			} else {
+				// Container has escaped its immediate parallel parent
+				current = current->pending.parent;
+				continue;
 			}
 		}
 
-		current = current->parent;
+		// We found a suitable ancestor, the loop will end
+		ancestor = current;
 	}
 
-	// Maybe rejigger the workspace
-	struct sway_workspace *ws = container->workspace;
-	if (ws) {
-		if (!is_parallel(ws->layout, move_dir)) {
-			workspace_rejigger(ws, container, move_dir);
-			return true;
-		} else if (ws->layout == L_TABBED || ws->layout == L_STACKED) {
-			workspace_rejigger(ws, container, move_dir);
-			return true;
+	if (target) {
+		// Container will move in with its cousin
+		container_move_to_container_from_direction(container,
+				target, move_dir);
+		return true;
+	} else if (!wrapped && !container->pending.parent->pending.parent &&
+			container->pending.parent->pending.children->length == 1) {
+		// Treat singleton children as if they are at workspace level like i3
+		// https://github.com/i3/i3/blob/1d9160f2d247dbaa83fb62f02fd7041dec767fc2/src/move.c#L367
+		return container_move_to_next_output(container,
+				ancestor->pending.workspace->output, move_dir);
+	} else {
+		// Container will be promoted
+		struct sway_container *old_parent = container->pending.parent;
+		if (ancestor->pending.parent) {
+			// Container will move in with its parent
+			container_insert_child(ancestor->pending.parent, container,
+					index + (offs < 0 ? 0 : 1));
+		} else {
+			// Container will move to workspace level,
+			// may be re-split by workspace_layout
+			workspace_insert_tiling(ancestor->pending.workspace, container,
+					index + (offs < 0 ? 0 : 1));
 		}
-
-		// Try adjacent output
-		struct sway_output *output =
-			output_get_in_direction(container->workspace->output, move_dir);
-		if (output) {
-			struct sway_workspace *ws = output_get_active_workspace(output);
-			if (!sway_assert(ws, "Expected output to have a workspace")) {
-				return false;
-			}
-			container_move_to_workspace_from_direction(container, ws, move_dir);
-			return true;
+		ancestor->pending.height = ancestor->pending.width = 0;
+		ancestor->height_fraction = ancestor->width_fraction = 0;
+		if (old_parent) {
+			container_reap_empty(old_parent);
 		}
-		sway_log(SWAY_DEBUG, "Hit edge of output, nowhere else to go");
+		workspace_squash(container->pending.workspace);
+		return true;
 	}
-	return false;
 }
 
 static struct cmd_results *cmd_move_to_scratchpad(void);
@@ -416,14 +427,14 @@ static struct cmd_results *cmd_move_container(bool no_auto_back_and_forth,
 		container = workspace_wrap_children(workspace);
 	}
 
-	if (container->fullscreen_mode == FULLSCREEN_GLOBAL) {
+	if (container->pending.fullscreen_mode == FULLSCREEN_GLOBAL) {
 		return cmd_results_new(CMD_FAILURE,
 				"Can't move fullscreen global container");
 	}
 
 	struct sway_seat *seat = config->handler_context.seat;
-	struct sway_container *old_parent = container->parent;
-	struct sway_workspace *old_ws = container->workspace;
+	struct sway_container *old_parent = container->pending.parent;
+	struct sway_workspace *old_ws = container->pending.workspace;
 	struct sway_output *old_output = old_ws ? old_ws->output : NULL;
 	struct sway_node *destination = NULL;
 
@@ -480,7 +491,7 @@ static struct cmd_results *cmd_move_container(bool no_auto_back_and_forth,
 			// We have to create the workspace, but if the container is
 			// sticky and the workspace is going to be created on the same
 			// output, we'll bail out first.
-			if (container->is_sticky && container_is_floating_or_child(container)) {
+			if (container_is_sticky_or_child(container)) {
 				struct sway_output *new_output =
 					workspace_get_initial_output(ws_name);
 				if (old_output == new_output) {
@@ -497,7 +508,7 @@ static struct cmd_results *cmd_move_container(bool no_auto_back_and_forth,
 		destination = dst ? &dst->node : &ws->node;
 	} else if (strcasecmp(argv[0], "output") == 0) {
 		struct sway_output *new_output = output_in_direction(argv[1],
-				old_output, container->x, container->y);
+				old_output, container->pending.x, container->pending.y);
 		if (!new_output) {
 			return cmd_results_new(CMD_FAILURE,
 				"Can't find output with name/direction '%s'", argv[1]);
@@ -519,8 +530,8 @@ static struct cmd_results *cmd_move_container(bool no_auto_back_and_forth,
 		return cmd_move_to_scratchpad();
 	}
 
-	if (container->is_sticky && container_is_floating_or_child(container) &&
-			old_output && node_has_ancestor(destination, &old_output->node)) {
+	if (container_is_sticky_or_child(container) && old_output &&
+			node_has_ancestor(destination, &old_output->node)) {
 		return cmd_results_new(CMD_FAILURE, "Can't move sticky "
 				"container to another workspace on the same output");
 	}
@@ -535,7 +546,8 @@ static struct cmd_results *cmd_move_container(bool no_auto_back_and_forth,
 	struct sway_node *focus = seat_get_focus(seat);
 
 	// move container
-	if (container->scratchpad) {
+	if (container_is_scratchpad_hidden_or_child(container)) {
+		container_detach(container);
 		root_scratchpad_show(container);
 	}
 	switch (destination->type) {
@@ -694,12 +706,12 @@ static struct cmd_results *cmd_move_in_direction(
 				"Cannot move workspaces in a direction");
 	}
 	if (container_is_floating(container)) {
-		if (container->fullscreen_mode) {
+		if (container->pending.fullscreen_mode) {
 			return cmd_results_new(CMD_FAILURE,
 					"Cannot move fullscreen floating container");
 		}
-		double lx = container->x;
-		double ly = container->y;
+		double lx = container->pending.x;
+		double ly = container->pending.y;
 		switch (direction) {
 		case WLR_DIRECTION_LEFT:
 			lx -= move_amt;
@@ -717,14 +729,22 @@ static struct cmd_results *cmd_move_in_direction(
 		container_floating_move_to(container, lx, ly);
 		return cmd_results_new(CMD_SUCCESS, NULL);
 	}
-	struct sway_workspace *old_ws = container->workspace;
+	struct sway_workspace *old_ws = container->pending.workspace;
+	struct sway_container *old_parent = container->pending.parent;
 
 	if (!container_move_in_direction(container, direction)) {
 		// Container didn't move
 		return cmd_results_new(CMD_SUCCESS, NULL);
 	}
 
-	struct sway_workspace *new_ws = container->workspace;
+	// clean-up, destroying parents if the container was the last child
+	if (old_parent) {
+		container_reap_empty(old_parent);
+	} else if (old_ws) {
+		workspace_consider_destroy(old_ws);
+	}
+
+	struct sway_workspace *new_ws = container->pending.workspace;
 
 	if (root->fullscreen_global) {
 		arrange_root();
@@ -753,6 +773,38 @@ static struct cmd_results *cmd_move_in_direction(
 	return cmd_results_new(CMD_SUCCESS, NULL);
 }
 
+static struct cmd_results *cmd_move_to_position_pointer(
+		struct sway_container *container) {
+	struct sway_seat *seat = config->handler_context.seat;
+	if (!seat->cursor) {
+		return cmd_results_new(CMD_FAILURE, "No cursor device");
+	}
+	struct wlr_cursor *cursor = seat->cursor->cursor;
+	/* Determine where to put the window. */
+	double lx = cursor->x - container->pending.width / 2;
+	double ly = cursor->y - container->pending.height / 2;
+
+	/* Correct target coordinates to be in bounds (on screen). */
+	struct wlr_output *output = wlr_output_layout_output_at(
+			root->output_layout, cursor->x, cursor->y);
+	if (output) {
+		struct wlr_box *box =
+				wlr_output_layout_get_box(root->output_layout, output);
+		lx = fmax(lx, box->x);
+		ly = fmax(ly, box->y);
+		if (lx + container->pending.width > box->x + box->width) {
+			lx = box->x + box->width - container->pending.width;
+		}
+		if (ly + container->pending.height > box->y + box->height) {
+			ly = box->y + box->height - container->pending.height;
+		}
+	}
+
+	/* Actually move the container. */
+	container_floating_move_to(container, lx, ly);
+	return cmd_results_new(CMD_SUCCESS, NULL);
+}
+
 static const char expected_position_syntax[] =
 	"Expected 'move [absolute] position <x> [px] <y> [px]' or "
 	"'move [absolute] position center' or "
@@ -790,27 +842,20 @@ static struct cmd_results *cmd_move_to_position(int argc, char **argv) {
 		if (absolute) {
 			return cmd_results_new(CMD_INVALID, expected_position_syntax);
 		}
-		struct sway_seat *seat = config->handler_context.seat;
-		if (!seat->cursor) {
-			return cmd_results_new(CMD_FAILURE, "No cursor device");
-		}
-		double lx = seat->cursor->cursor->x - container->width / 2;
-		double ly = seat->cursor->cursor->y - container->height / 2;
-		container_floating_move_to(container, lx, ly);
-		return cmd_results_new(CMD_SUCCESS, NULL);
+		return cmd_move_to_position_pointer(container);
 	} else if (strcmp(argv[0], "center") == 0) {
 		double lx, ly;
 		if (absolute) {
-			lx = root->x + (root->width - container->width) / 2;
-			ly = root->y + (root->height - container->height) / 2;
+			lx = root->x + (root->width - container->pending.width) / 2;
+			ly = root->y + (root->height - container->pending.height) / 2;
 		} else {
-			struct sway_workspace *ws = container->workspace;
+			struct sway_workspace *ws = container->pending.workspace;
 			if (!ws) {
 				struct sway_seat *seat = config->handler_context.seat;
 				ws = seat_get_focused_workspace(seat);
 			}
-			lx = ws->x + (ws->width - container->width) / 2;
-			ly = ws->y + (ws->height - container->height) / 2;
+			lx = ws->x + (ws->width - container->pending.width) / 2;
+			ly = ws->y + (ws->height - container->pending.height) / 2;
 		}
 		container_floating_move_to(container, lx, ly);
 		return cmd_results_new(CMD_SUCCESS, NULL);
@@ -820,37 +865,81 @@ static struct cmd_results *cmd_move_to_position(int argc, char **argv) {
 		return cmd_results_new(CMD_FAILURE, expected_position_syntax);
 	}
 
-	double lx, ly;
-	char *inv;
-	lx = (double)strtol(argv[0], &inv, 10);
-	if (*inv != '\0' && strcasecmp(inv, "px") != 0) {
-		return cmd_results_new(CMD_FAILURE, "Invalid position specified");
-	}
-	if (strcmp(argv[1], "px") == 0) {
-		--argc;
-		++argv;
+	struct movement_amount lx = { .amount = 0, .unit = MOVEMENT_UNIT_INVALID };
+	// X direction
+	int num_consumed_args = parse_movement_amount(argc, argv, &lx);
+	argc -= num_consumed_args;
+	argv += num_consumed_args;
+	if (lx.unit == MOVEMENT_UNIT_INVALID) {
+		return cmd_results_new(CMD_INVALID, "Invalid x position specified");
 	}
 
-	if (argc > 3) {
-		return cmd_results_new(CMD_FAILURE, expected_position_syntax);
+	struct movement_amount ly = { .amount = 0, .unit = MOVEMENT_UNIT_INVALID };
+	// Y direction
+	num_consumed_args = parse_movement_amount(argc, argv, &ly);
+	argc -= num_consumed_args;
+	argv += num_consumed_args;
+	if (argc > 0) {
+		return cmd_results_new(CMD_INVALID, expected_position_syntax);
+	}
+	if (ly.unit == MOVEMENT_UNIT_INVALID) {
+		return cmd_results_new(CMD_INVALID, "Invalid y position specified");
 	}
 
-	ly = (double)strtol(argv[1], &inv, 10);
-	if ((*inv != '\0' && strcasecmp(inv, "px") != 0) ||
-			(argc == 3 && strcmp(argv[2], "px") != 0)) {
-		return cmd_results_new(CMD_FAILURE, "Invalid position specified");
+	struct sway_workspace *ws = container->pending.workspace;
+	if (!ws) {
+		struct sway_seat *seat = config->handler_context.seat;
+		ws = seat_get_focused_workspace(seat);
 	}
 
-	if (!absolute) {
-		struct sway_workspace *ws = container->workspace;
-		if (!ws) {
-			struct sway_seat *seat = config->handler_context.seat;
-			ws = seat_get_focused_workspace(seat);
+	switch (lx.unit) {
+	case MOVEMENT_UNIT_PPT:
+		if (container_is_scratchpad_hidden(container)) {
+			return cmd_results_new(CMD_FAILURE,
+					"Cannot move a hidden scratchpad container by ppt");
 		}
-		lx += ws->x;
-		ly += ws->y;
+		if (absolute) {
+			return cmd_results_new(CMD_FAILURE,
+					"Cannot move to absolute positions by ppt");
+		}
+		// Convert to px
+		lx.amount = ws->width * lx.amount / 100;
+		lx.unit = MOVEMENT_UNIT_PX;
+		// Falls through
+	case MOVEMENT_UNIT_PX:
+	case MOVEMENT_UNIT_DEFAULT:
+		break;
+	case MOVEMENT_UNIT_INVALID:
+		sway_assert(false, "invalid x unit");
+		break;
 	}
-	container_floating_move_to(container, lx, ly);
+
+	switch (ly.unit) {
+	case MOVEMENT_UNIT_PPT:
+		if (container_is_scratchpad_hidden(container)) {
+			return cmd_results_new(CMD_FAILURE,
+					"Cannot move a hidden scratchpad container by ppt");
+		}
+		if (absolute) {
+			return cmd_results_new(CMD_FAILURE,
+					"Cannot move to absolute positions by ppt");
+		}
+		// Convert to px
+		ly.amount = ws->height * ly.amount / 100;
+		ly.unit = MOVEMENT_UNIT_PX;
+		// Falls through
+	case MOVEMENT_UNIT_PX:
+	case MOVEMENT_UNIT_DEFAULT:
+		break;
+	case MOVEMENT_UNIT_INVALID:
+		sway_assert(false, "invalid y unit");
+		break;
+	}
+	if (!absolute) {
+		lx.amount += ws->x;
+		ly.amount += ws->y;
+	}
+	container_floating_move_to(container, lx.amount, ly.amount);
 	return cmd_results_new(CMD_SUCCESS, NULL);
 }
 
@@ -871,14 +960,14 @@ static struct cmd_results *cmd_move_to_scratchpad(void) {
 	// If the container is in a floating split container,
 	// operate on the split container instead of the child.
 	if (container_is_floating_or_child(con)) {
-		while (con->parent) {
-			con = con->parent;
+		while (con->pending.parent) {
+			con = con->pending.parent;
 		}
 	}
 
 	if (!con->scratchpad) {
 		root_scratchpad_add_container(con, NULL);
-	} else if (con->workspace) {
+	} else if (con->pending.workspace) {
 		root_scratchpad_hide(con);
 	}
 	return cmd_results_new(CMD_SUCCESS, NULL);
