@@ -32,6 +32,7 @@
 struct render_data {
 	pixman_region32_t *damage;
 	float alpha;
+	struct sway_container *container;
 };
 
 /**
@@ -149,15 +150,23 @@ static void render_surface_iterator(struct sway_output *output, struct sway_view
 	struct wlr_fbox src_box;
 	wlr_surface_get_buffer_source_box(surface, &src_box);
 
-	struct wlr_box dst_box = *_box;
-	scale_box(&dst_box, wlr_output->scale);
+	struct wlr_box proj_box = *_box;
+	scale_box(&proj_box, wlr_output->scale);
 
 	float matrix[9];
 	enum wl_output_transform transform =
 		wlr_output_transform_invert(surface->current.transform);
-	wlr_matrix_project_box(matrix, &dst_box, transform, rotation,
+	wlr_matrix_project_box(matrix, &proj_box, transform, rotation,
 		wlr_output->transform_matrix);
 
+	struct wlr_box dst_box = *_box;
+	struct sway_container *container = data->container;
+	if (container != NULL) {
+		dst_box.width = fmin(dst_box.width, container->current.content_width - surface->sx);
+		dst_box.height = fmin(dst_box.height, container->current.content_height - surface->sy);
+	}
+	scale_box(&dst_box, wlr_output->scale);
+
 	render_texture(wlr_output, output_damage, texture,
 		&src_box, &dst_box, matrix, alpha);
 
@@ -171,7 +180,7 @@ static void render_layer_toplevel(struct sway_output *output,
 		.damage = damage,
 		.alpha = 1.0f,
 	};
-	output_layer_for_each_surface_toplevel(output, layer_surfaces,
+	output_layer_for_each_toplevel_surface(output, layer_surfaces,
 		render_surface_iterator, &data);
 }
 
@@ -181,7 +190,7 @@ static void render_layer_popups(struct sway_output *output,
 		.damage = damage,
 		.alpha = 1.0f,
 	};
-	output_layer_for_each_surface_popup(output, layer_surfaces,
+	output_layer_for_each_popup_surface(output, layer_surfaces,
 		render_surface_iterator, &data);
 }
 
@@ -256,6 +265,9 @@ static void render_view_toplevels(struct sway_view *view,
 		.damage = damage,
 		.alpha = alpha,
 	};
+	if (!container_is_current_floating(view->container)) {
+		data.container = view->container;
+	}
 	// Render all toplevels without descending into popups
 	double ox = view->container->surface_x -
 		output->lx - view->geometry.x;
@@ -265,24 +277,14 @@ static void render_view_toplevels(struct sway_view *view,
 			render_surface_iterator, &data);
 }
 
-static void render_popup_iterator(struct sway_output *output, struct sway_view *view,
-		struct wlr_surface *surface, struct wlr_box *box, float rotation,
-		void *data) {
-	// Render this popup's surface
-	render_surface_iterator(output, view, surface, box, rotation, data);
-
-	// Render this popup's child toplevels
-	output_surface_for_each_surface(output, surface, box->x, box->y,
-			render_surface_iterator, data);
-}
-
 static void render_view_popups(struct sway_view *view,
 		struct sway_output *output, pixman_region32_t *damage, float alpha) {
 	struct render_data data = {
 		.damage = damage,
 		.alpha = alpha,
 	};
-	output_view_for_each_popup(output, view, render_popup_iterator, &data);
+	output_view_for_each_popup_surface(output, view,
+		render_surface_iterator, &data);
 }
 
 static void render_saved_view(struct sway_view *view,
@@ -292,17 +294,18 @@ static void render_saved_view(struct sway_view *view,
 	if (wl_list_empty(&view->saved_buffers)) {
 		return;
 	}
+
+	bool floating = container_is_current_floating(view->container);
+
 	struct sway_saved_buffer *saved_buf;
 	wl_list_for_each(saved_buf, &view->saved_buffers, link) {
 		if (!saved_buf->buffer->texture) {
 			continue;
 		}
 
-		struct wlr_box box = {
-			.x = view->container->surface_x - output->lx -
-				view->saved_geometry.x + saved_buf->x,
-			.y = view->container->surface_y - output->ly -
-				view->saved_geometry.y + saved_buf->y,
+		struct wlr_box proj_box = {
+			.x = saved_buf->x - view->saved_geometry.x - output->lx,
+			.y = saved_buf->y - view->saved_geometry.y - output->ly,
 			.width = saved_buf->width,
 			.height = saved_buf->height,
 		};
@@ -313,20 +316,31 @@ static void render_saved_view(struct sway_view *view,
 		};
 
 		struct wlr_box intersection;
-		bool intersects = wlr_box_intersection(&intersection, &output_box, &box);
+		bool intersects = wlr_box_intersection(&intersection, &output_box, &proj_box);
 		if (!intersects) {
 			continue;
 		}
 
-		scale_box(&box, wlr_output->scale);
+		struct wlr_box dst_box = proj_box;
+		scale_box(&proj_box, wlr_output->scale);
 
 		float matrix[9];
 		enum wl_output_transform transform = wlr_output_transform_invert(saved_buf->transform);
-		wlr_matrix_project_box(matrix, &box, transform, 0,
+		wlr_matrix_project_box(matrix, &proj_box, transform, 0,
 			wlr_output->transform_matrix);
 
+		if (!floating) {
+			dst_box.width = fmin(dst_box.width,
+					view->container->current.content_width -
+					(saved_buf->x - view->container->current.content_x));
+			dst_box.height = fmin(dst_box.height,
+					view->container->current.content_height -
+					(saved_buf->y - view->container->current.content_y));
+		}
+		scale_box(&dst_box, wlr_output->scale);
+
 		render_texture(wlr_output, damage, saved_buf->buffer->texture,
-			&saved_buf->source_box, &box, matrix, alpha);
+			&saved_buf->source_box, &dst_box, matrix, alpha);
 	}
 
 	// FIXME: we should set the surface that this saved buffer originates from
@@ -358,8 +372,8 @@ static void render_view(struct sway_output *output, pixman_region32_t *damage,
 	if (state->border_left) {
 		memcpy(&color, colors->child_border, sizeof(float) * 4);
 		premultiply_alpha(color, con->alpha);
-		box.x = state->x;
-		box.y = state->content_y;
+		box.x = floor(state->x);
+		box.y = floor(state->content_y);
 		box.width = state->border_thickness;
 		box.height = state->content_height;
 		scale_box(&box, output_scale);
@@ -371,14 +385,14 @@ static void render_view(struct sway_output *output, pixman_region32_t *damage,
 		container_current_parent_layout(con);
 
 	if (state->border_right) {
-		if (con->current.parent && siblings->length == 1 && layout == L_HORIZ) {
+		if (!container_is_current_floating(con) && siblings->length == 1 && layout == L_HORIZ) {
 			memcpy(&color, colors->indicator, sizeof(float) * 4);
 		} else {
 			memcpy(&color, colors->child_border, sizeof(float) * 4);
 		}
 		premultiply_alpha(color, con->alpha);
-		box.x = state->content_x + state->content_width;
-		box.y = state->content_y;
+		box.x = floor(state->content_x + state->content_width);
+		box.y = floor(state->content_y);
 		box.width = state->border_thickness;
 		box.height = state->content_height;
 		scale_box(&box, output_scale);
@@ -386,14 +400,14 @@ static void render_view(struct sway_output *output, pixman_region32_t *damage,
 	}
 
 	if (state->border_bottom) {
-		if (con->current.parent && siblings->length == 1 && layout == L_VERT) {
+		if (!container_is_current_floating(con) && siblings->length == 1 && layout == L_VERT) {
 			memcpy(&color, colors->indicator, sizeof(float) * 4);
 		} else {
 			memcpy(&color, colors->child_border, sizeof(float) * 4);
 		}
 		premultiply_alpha(color, con->alpha);
-		box.x = state->x;
-		box.y = state->content_y + state->content_height;
+		box.x = floor(state->x);
+		box.y = floor(state->content_y + state->content_height);
 		box.width = state->width;
 		box.height = state->border_thickness;
 		scale_box(&box, output_scale);
@@ -530,6 +544,14 @@ static void render_titlebar(struct sway_output *output,
 		struct wlr_box texture_box;
 		wlr_texture_get_size(title_texture,
 			&texture_box.width, &texture_box.height);
+
+		// The effective output may be NULL when con is not on any output.
+		// This can happen because we render all children of containers,
+		// even those that are out of the bounds of any output.
+		struct sway_output *effective = container_get_effective_output(con);
+		float title_scale = effective ? effective->wlr_output->scale : output_scale;
+		texture_box.width = texture_box.width * output_scale / title_scale;
+		texture_box.height = texture_box.height * output_scale / title_scale;
 		ob_title_width = texture_box.width;
 
 		// The title texture might be shorter than the config->font_height,
@@ -670,8 +692,8 @@ static void render_top_border(struct sway_output *output,
 	// Child border - top edge
 	memcpy(&color, colors->child_border, sizeof(float) * 4);
 	premultiply_alpha(color, con->alpha);
-	box.x = state->x;
-	box.y = state->y;
+	box.x = floor(state->x);
+	box.y = floor(state->y);
 	box.width = state->width;
 	box.height = state->border_thickness;
 	scale_box(&box, output_scale);
@@ -726,8 +748,8 @@ static void render_containers_linear(struct sway_output *output,
 			}
 
 			if (state->border == B_NORMAL) {
-				render_titlebar(output, damage, child, state->x,
-						state->y, state->width, colors,
+				render_titlebar(output, damage, child, floor(state->x),
+						floor(state->y), state->width, colors,
 						title_texture, marks_texture);
 			} else if (state->border == B_PIXEL) {
 				render_top_border(output, damage, child, colors);
@@ -781,7 +803,7 @@ static void render_containers_tabbed(struct sway_output *output,
 			marks_texture = child->marks_unfocused;
 		}
 
-		int x = cstate->x + tab_width * i;
+		int x = floor(cstate->x + tab_width * i);
 
 		// Make last tab use the remaining width of the parent
 		if (i == parent->children->length - 1) {
@@ -894,8 +916,8 @@ static void render_container(struct sway_output *output,
 	struct parent_data data = {
 		.layout = con->current.layout,
 		.box = {
-			.x = con->current.x,
-			.y = con->current.y,
+			.x = floor(con->current.x),
+			.y = floor(con->current.y),
 			.width = con->current.width,
 			.height = con->current.height,
 		},
@@ -911,8 +933,8 @@ static void render_workspace(struct sway_output *output,
 	struct parent_data data = {
 		.layout = ws->current.layout,
 		.box = {
-			.x = ws->current.x,
-			.y = ws->current.y,
+			.x = floor(ws->current.x),
+			.y = floor(ws->current.y),
 			.width = ws->current.width,
 			.height = ws->current.height,
 		},
@@ -946,8 +968,8 @@ static void render_floating_container(struct sway_output *soutput,
 		}
 
 		if (con->current.border == B_NORMAL) {
-			render_titlebar(soutput, damage, con, con->current.x,
-					con->current.y, con->current.width, colors,
+			render_titlebar(soutput, damage, con, floor(con->current.x),
+					floor(con->current.y), con->current.width, colors,
 					title_texture, marks_texture);
 		} else if (con->current.border == B_PIXEL) {
 			render_top_border(soutput, damage, con, colors);
@@ -969,7 +991,7 @@ static void render_floating(struct sway_output *soutput,
 			}
 			for (int k = 0; k < ws->current.floating->length; ++k) {
 				struct sway_container *floater = ws->current.floating->items[k];
-				if (floater->fullscreen_mode != FULLSCREEN_NONE) {
+				if (floater->current.fullscreen_mode != FULLSCREEN_NONE) {
 					continue;
 				}
 				render_floating_container(soutput, damage, floater);
