@@ -8,7 +8,11 @@
 #include <wlr/types/wlr_box.h>
 #include <wlr/types/wlr_cursor.h>
 #include <wlr/types/wlr_idle.h>
+#include <wlr/types/wlr_pointer.h>
+#include <wlr/types/wlr_touch.h>
 #include <wlr/types/wlr_tablet_v2.h>
+#include <wlr/types/wlr_tablet_pad.h>
+#include <wlr/types/wlr_tablet_tool.h>
 #include <wlr/types/wlr_xcursor_manager.h>
 #include <wlr/util/region.h>
 #include "config.h"
@@ -16,7 +20,6 @@
 #include "util.h"
 #include "sway/commands.h"
 #include "sway/desktop.h"
-#include "sway/desktop/transaction.h"
 #include "sway/input/cursor.h"
 #include "sway/input/keyboard.h"
 #include "sway/input/tablet.h"
@@ -253,6 +256,32 @@ int cursor_get_timeout(struct sway_cursor *cursor) {
 	return timeout;
 }
 
+void cursor_notify_key_press(struct sway_cursor *cursor) {
+	if (cursor->hidden) {
+		return;
+	}
+
+	if (cursor->hide_when_typing == HIDE_WHEN_TYPING_DEFAULT) {
+		// No cached value, need to lookup in the seat_config
+		const struct seat_config *seat_config = seat_get_config(cursor->seat);
+		if (!seat_config) {
+			seat_config = seat_get_config_by_name("*");
+			if (!seat_config) {
+				return;
+			}
+		}
+		cursor->hide_when_typing = seat_config->hide_cursor_when_typing;
+		// The default is currently disabled
+		if (cursor->hide_when_typing == HIDE_WHEN_TYPING_DEFAULT) {
+			cursor->hide_when_typing = HIDE_WHEN_TYPING_DISABLE;
+		}
+	}
+
+	if (cursor->hide_when_typing == HIDE_WHEN_TYPING_ENABLE) {
+		cursor_hide(cursor);
+	}
+}
+
 static enum sway_input_idle_source idle_source_from_device(
 		struct wlr_input_device *device) {
 	switch (device->type) {
@@ -273,19 +302,28 @@ static enum sway_input_idle_source idle_source_from_device(
 	abort();
 }
 
-void cursor_handle_activity(struct sway_cursor *cursor,
-		struct wlr_input_device *device) {
+void cursor_handle_activity_from_idle_source(struct sway_cursor *cursor,
+		enum sway_input_idle_source idle_source) {
 	wl_event_source_timer_update(
 			cursor->hide_source, cursor_get_timeout(cursor));
 
-	enum sway_input_idle_source idle_source = idle_source_from_device(device);
 	seat_idle_notify_activity(cursor->seat, idle_source);
-	if (cursor->hidden && idle_source != IDLE_SOURCE_TOUCH) {
+	if (idle_source != IDLE_SOURCE_TOUCH) {
 		cursor_unhide(cursor);
 	}
 }
 
+void cursor_handle_activity_from_device(struct sway_cursor *cursor,
+		struct wlr_input_device *device) {
+	enum sway_input_idle_source idle_source = idle_source_from_device(device);
+	cursor_handle_activity_from_idle_source(cursor, idle_source);
+}
+
 void cursor_unhide(struct sway_cursor *cursor) {
+	if (!cursor->hidden) {
+		return;
+	}
+
 	cursor->hidden = false;
 	if (cursor->image_surface) {
 		cursor_set_image_surface(cursor,
@@ -299,6 +337,7 @@ void cursor_unhide(struct sway_cursor *cursor) {
 		cursor_set_image(cursor, image, cursor->image_client);
 	}
 	cursor_rebase(cursor);
+	wl_event_source_timer_update(cursor->hide_source, cursor_get_timeout(cursor));
 }
 
 static void pointer_motion(struct sway_cursor *cursor, uint32_t time_msec,
@@ -332,18 +371,17 @@ static void pointer_motion(struct sway_cursor *cursor, uint32_t time_msec,
 
 	wlr_cursor_move(cursor->cursor, device, dx, dy);
 
-	seatop_pointer_motion(cursor->seat, time_msec, dx, dy);
+	seatop_pointer_motion(cursor->seat, time_msec);
 }
 
 static void handle_pointer_motion_relative(
 		struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, motion);
 	struct wlr_event_pointer_motion *e = data;
-	cursor_handle_activity(cursor, e->device);
+	cursor_handle_activity_from_device(cursor, e->device);
 
 	pointer_motion(cursor, e->time_msec, e->device, e->delta_x, e->delta_y,
 			e->unaccel_dx, e->unaccel_dy);
-	transaction_commit_dirty();
 }
 
 static void handle_pointer_motion_absolute(
@@ -351,7 +389,7 @@ static void handle_pointer_motion_absolute(
 	struct sway_cursor *cursor =
 		wl_container_of(listener, cursor, motion_absolute);
 	struct wlr_event_pointer_motion_absolute *event = data;
-	cursor_handle_activity(cursor, event->device);
+	cursor_handle_activity_from_device(cursor, event->device);
 
 	double lx, ly;
 	wlr_cursor_absolute_to_layout_coords(cursor->cursor, event->device,
@@ -361,7 +399,6 @@ static void handle_pointer_motion_absolute(
 	double dy = ly - cursor->cursor->y;
 
 	pointer_motion(cursor, event->time_msec, event->device, dx, dy, dx, dy);
-	transaction_commit_dirty();
 }
 
 void dispatch_cursor_button(struct sway_cursor *cursor,
@@ -377,7 +414,6 @@ void dispatch_cursor_button(struct sway_cursor *cursor,
 static void handle_pointer_button(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, button);
 	struct wlr_event_pointer_button *event = data;
-	cursor_handle_activity(cursor, event->device);
 
 	if (event->state == WLR_BUTTON_PRESSED) {
 		cursor->pressed_button_count++;
@@ -389,9 +425,9 @@ static void handle_pointer_button(struct wl_listener *listener, void *data) {
 		}
 	}
 
+	cursor_handle_activity_from_device(cursor, event->device);
 	dispatch_cursor_button(cursor, event->device,
 			event->time_msec, event->button, event->state);
-	transaction_commit_dirty();
 }
 
 void dispatch_cursor_axis(struct sway_cursor *cursor,
@@ -402,9 +438,8 @@ void dispatch_cursor_axis(struct sway_cursor *cursor,
 static void handle_pointer_axis(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, axis);
 	struct wlr_event_pointer_axis *event = data;
-	cursor_handle_activity(cursor, event->device);
+	cursor_handle_activity_from_device(cursor, event->device);
 	dispatch_cursor_axis(cursor, event);
-	transaction_commit_dirty();
 }
 
 static void handle_pointer_frame(struct wl_listener *listener, void *data) {
@@ -415,7 +450,7 @@ static void handle_pointer_frame(struct wl_listener *listener, void *data) {
 static void handle_touch_down(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, touch_down);
 	struct wlr_event_touch_down *event = data;
-	cursor_handle_activity(cursor, event->device);
+	cursor_handle_activity_from_device(cursor, event->device);
 	cursor_hide(cursor);
 
 	struct sway_seat *seat = cursor->seat;
@@ -455,14 +490,13 @@ static void handle_touch_down(struct wl_listener *listener, void *data) {
 		dispatch_cursor_button(cursor, event->device, event->time_msec,
 				BTN_LEFT, WLR_BUTTON_PRESSED);
 		wlr_seat_pointer_notify_frame(wlr_seat);
-		transaction_commit_dirty();
 	}
 }
 
 static void handle_touch_up(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, touch_up);
 	struct wlr_event_touch_up *event = data;
-	cursor_handle_activity(cursor, event->device);
+	cursor_handle_activity_from_device(cursor, event->device);
 
 	struct wlr_seat *wlr_seat = cursor->seat->wlr_seat;
 
@@ -472,7 +506,6 @@ static void handle_touch_up(struct wl_listener *listener, void *data) {
 			dispatch_cursor_button(cursor, event->device, event->time_msec,
 					BTN_LEFT, WLR_BUTTON_RELEASED);
 			wlr_seat_pointer_notify_frame(wlr_seat);
-			transaction_commit_dirty();
 		}
 	} else {
 		wlr_seat_touch_notify_up(wlr_seat, event->time_msec, event->touch_id);
@@ -483,7 +516,7 @@ static void handle_touch_motion(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor =
 		wl_container_of(listener, cursor, touch_motion);
 	struct wlr_event_touch_motion *event = data;
-	cursor_handle_activity(cursor, event->device);
+	cursor_handle_activity_from_device(cursor, event->device);
 
 	struct sway_seat *seat = cursor->seat;
 	struct wlr_seat *wlr_seat = seat->wlr_seat;
@@ -513,7 +546,6 @@ static void handle_touch_motion(struct wl_listener *listener, void *data) {
 			dx = lx - cursor->cursor->x;
 			dy = ly - cursor->cursor->y;
 			pointer_motion(cursor, event->time_msec, event->device, dx, dy, dx, dy);
-			transaction_commit_dirty();
 		}
 	} else if (surface) {
 		wlr_seat_touch_notify_motion(wlr_seat, event->time_msec,
@@ -566,13 +598,14 @@ static void handle_tablet_tool_position(struct sway_cursor *cursor,
 			ic->mapped_from_region, &x, &y);
 	}
 
-	switch (tool->tablet_v2_tool->wlr_tool->type) {
-	case WLR_TABLET_TOOL_TYPE_MOUSE:
-		wlr_cursor_move(cursor->cursor, input_device->wlr_device, dx, dy);
-		break;
-	default:
+	switch (tool->mode) {
+	case SWAY_TABLET_TOOL_MODE_ABSOLUTE:
 		wlr_cursor_warp_absolute(cursor->cursor, input_device->wlr_device,
 			change_x ? x : NAN, change_y ? y : NAN);
+		break;
+	case SWAY_TABLET_TOOL_MODE_RELATIVE:
+		wlr_cursor_move(cursor->cursor, input_device->wlr_device, dx, dy);
+		break;
 	}
 
 	double sx, sy;
@@ -593,19 +626,17 @@ static void handle_tablet_tool_position(struct sway_cursor *cursor,
 	if (!cursor->simulating_pointer_from_tool_tip &&
 			((surface && wlr_surface_accepts_tablet_v2(tablet->tablet_v2, surface)) ||
 				wlr_tablet_tool_v2_has_implicit_grab(tool->tablet_v2_tool))) {
-		seatop_tablet_tool_motion(seat, tool, time_msec, dx, dy);
+		seatop_tablet_tool_motion(seat, tool, time_msec);
 	} else {
 		wlr_tablet_v2_tablet_tool_notify_proximity_out(tool->tablet_v2_tool);
 		pointer_motion(cursor, time_msec, input_device->wlr_device, dx, dy, dx, dy);
 	}
-
-	transaction_commit_dirty();
 }
 
 static void handle_tool_axis(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, tool_axis);
 	struct wlr_event_tablet_tool_axis *event = data;
-	cursor_handle_activity(cursor, event->device);
+	cursor_handle_activity_from_device(cursor, event->device);
 
 	struct sway_tablet_tool *sway_tool = event->tool->data;
 	if (!sway_tool) {
@@ -661,7 +692,7 @@ static void handle_tool_axis(struct wl_listener *listener, void *data) {
 static void handle_tool_tip(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, tool_tip);
 	struct wlr_event_tablet_tool_tip *event = data;
-	cursor_handle_activity(cursor, event->device);
+	cursor_handle_activity_from_device(cursor, event->device);
 
 	struct sway_tablet_tool *sway_tool = event->tool->data;
 	struct wlr_tablet_v2_tablet *tablet_v2 = sway_tool->tablet->tablet_v2;
@@ -679,7 +710,6 @@ static void handle_tool_tip(struct wl_listener *listener, void *data) {
 		dispatch_cursor_button(cursor, event->device, event->time_msec,
 			BTN_LEFT, WLR_BUTTON_RELEASED);
 		wlr_seat_pointer_notify_frame(cursor->seat->wlr_seat);
-		transaction_commit_dirty();
 	} else if (!surface || !wlr_surface_accepts_tablet_v2(tablet_v2, surface)) {
 		// If we started holding the tool tip down on a surface that accepts
 		// tablet v2, we should notify that surface if it gets released over a
@@ -692,7 +722,6 @@ static void handle_tool_tip(struct wl_listener *listener, void *data) {
 			dispatch_cursor_button(cursor, event->device, event->time_msec,
 				BTN_LEFT, WLR_BUTTON_PRESSED);
 			wlr_seat_pointer_notify_frame(cursor->seat->wlr_seat);
-			transaction_commit_dirty();
 		}
 	} else {
 		seatop_tablet_tool_tip(seat, sway_tool, event->time_msec, event->state);
@@ -714,7 +743,7 @@ static void handle_tool_proximity(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor =
 		wl_container_of(listener, cursor, tool_proximity);
 	struct wlr_event_tablet_tool_proximity *event = data;
-	cursor_handle_activity(cursor, event->device);
+	cursor_handle_activity_from_device(cursor, event->device);
 
 	struct wlr_tablet_tool *tool = event->tool;
 	if (!tool->data) {
@@ -744,7 +773,7 @@ static void handle_tool_proximity(struct wl_listener *listener, void *data) {
 static void handle_tool_button(struct wl_listener *listener, void *data) {
 	struct sway_cursor *cursor = wl_container_of(listener, cursor, tool_button);
 	struct wlr_event_tablet_tool_button *event = data;
-	cursor_handle_activity(cursor, event->device);
+	cursor_handle_activity_from_device(cursor, event->device);
 
 	struct sway_tablet_tool *sway_tool = event->tool->data;
 	if (!sway_tool) {
@@ -779,7 +808,6 @@ static void handle_tool_button(struct wl_listener *listener, void *data) {
 			break;
 		}
 		wlr_seat_pointer_notify_frame(cursor->seat->wlr_seat);
-		transaction_commit_dirty();
 		return;
 	}
 
@@ -796,8 +824,8 @@ static void check_constraint_region(struct sway_cursor *cursor) {
 
 		struct sway_container *con = view->container;
 
-		double sx = cursor->cursor->x - con->content_x + view->geometry.x;
-		double sy = cursor->cursor->y - con->content_y + view->geometry.y;
+		double sx = cursor->cursor->x - con->pending.content_x + view->geometry.x;
+		double sy = cursor->cursor->y - con->pending.content_y + view->geometry.y;
 
 		if (!pixman_region32_contains_point(region,
 				floor(sx), floor(sy), NULL)) {
@@ -808,8 +836,8 @@ static void check_constraint_region(struct sway_cursor *cursor) {
 				double sy = (boxes[0].y1 + boxes[0].y2) / 2.;
 
 				wlr_cursor_warp_closest(cursor->cursor, NULL,
-					sx + con->content_x - view->geometry.x,
-					sy + con->content_y - view->geometry.y);
+					sx + con->pending.content_x - view->geometry.x,
+					sy + con->pending.content_y - view->geometry.y);
 
 				cursor_rebase(cursor);
 			}
@@ -1113,25 +1141,27 @@ struct sway_cursor *sway_cursor_create(struct sway_seat *seat) {
 
 /**
  * Warps the cursor to the middle of the container argument.
- * Does nothing if the cursor is already inside the container.
- * If container is NULL, returns without doing anything.
+ * Does nothing if the cursor is already inside the container and `force` is
+ * false. If container is NULL, returns without doing anything.
  */
 void cursor_warp_to_container(struct sway_cursor *cursor,
-		struct sway_container *container) {
+		struct sway_container *container, bool force) {
 	if (!container) {
 		return;
 	}
 
 	struct wlr_box box;
 	container_get_box(container, &box);
-	if (wlr_box_contains_point(&box, cursor->cursor->x, cursor->cursor->y)) {
+	if (!force && wlr_box_contains_point(&box, cursor->cursor->x,
+			cursor->cursor->y)) {
 		return;
 	}
 
-	double x = container->x + container->width / 2.0;
-	double y = container->y + container->height / 2.0;
+	double x = container->pending.x + container->pending.width / 2.0;
+	double y = container->pending.y + container->pending.height / 2.0;
 
 	wlr_cursor_warp(cursor->cursor, NULL, x, y);
+	cursor_unhide(cursor);
 }
 
 /**
@@ -1148,6 +1178,7 @@ void cursor_warp_to_workspace(struct sway_cursor *cursor,
 	double y = workspace->y + workspace->height / 2.0;
 
 	wlr_cursor_warp(cursor->cursor, NULL, x, y);
+	cursor_unhide(cursor);
 }
 
 uint32_t get_mouse_bindsym(const char *name, char **error) {
@@ -1240,8 +1271,8 @@ static void warp_to_constraint_cursor_hint(struct sway_cursor *cursor) {
 		struct sway_view *view = view_from_wlr_surface(constraint->surface);
 		struct sway_container *con = view->container;
 
-		double lx = sx + con->content_x - view->geometry.x;
-		double ly = sy + con->content_y - view->geometry.y;
+		double lx = sx + con->pending.content_x - view->geometry.x;
+		double ly = sy + con->pending.content_y - view->geometry.y;
 
 		wlr_cursor_warp(cursor->cursor, NULL, lx, ly);
 
