@@ -48,7 +48,7 @@ struct sway_output *workspace_get_initial_output(const char *name) {
 	if (focus && focus->type == N_WORKSPACE) {
 		return focus->sway_workspace->output;
 	} else if (focus && focus->type == N_CONTAINER) {
-		return focus->sway_container->workspace->output;
+		return focus->sway_container->pending.workspace->output;
 	}
 	// Fallback to the first output or noop output for headless
 	return root->outputs->length ? root->outputs->items[0] : root->noop_output;
@@ -382,6 +382,148 @@ struct sway_workspace *workspace_by_name(const char *name) {
 	}
 }
 
+static int workspace_get_number(struct sway_workspace *workspace) {
+	char *endptr = NULL;
+	errno = 0;
+	long long n = strtoll(workspace->name, &endptr, 10);
+	if (errno != 0 || n > INT32_MAX || n < 0 || endptr == workspace->name) {
+		n = -1;
+	}
+	return n;
+}
+
+struct sway_workspace *workspace_prev(struct sway_workspace *workspace) {
+	int n = workspace_get_number(workspace);
+	struct sway_workspace *prev = NULL, *last = NULL, *other = NULL;
+	bool found = false;
+	if (n < 0) {
+		// Find the prev named workspace
+		int othern = -1;
+		for (int i = root->outputs->length - 1; i >= 0; i--) {
+			struct sway_output *output = root->outputs->items[i];
+			for (int j = output->workspaces->length - 1; j >= 0; j--) {
+				struct sway_workspace *ws = output->workspaces->items[j];
+				int wsn = workspace_get_number(ws);
+				if (!last) {
+					// The first workspace in reverse order
+					last = ws;
+				}
+				if (!other || (wsn >= 0 && wsn > othern)) {
+					// The last (greatest) numbered workspace.
+					other = ws;
+					othern = workspace_get_number(other);
+				}
+				if (ws == workspace) {
+					found = true;
+				} else if (wsn < 0 && found) {
+					// Found a non-numbered workspace before current
+					return ws;
+				}
+			}
+		}
+	} else {
+		// Find the prev numbered workspace
+		int prevn = -1, lastn = -1;
+		for (int i = root->outputs->length - 1; i >= 0; i--) {
+			struct sway_output *output = root->outputs->items[i];
+			for (int j = output->workspaces->length - 1; j >= 0; j--) {
+				struct sway_workspace *ws = output->workspaces->items[j];
+				int wsn = workspace_get_number(ws);
+				if (!last || (wsn >= 0 && wsn > lastn)) {
+					// The greatest numbered (or last) workspace
+					last = ws;
+					lastn = workspace_get_number(last);
+				}
+				if (!other && wsn < 0) {
+					// The last named workspace
+					other = ws;
+				}
+				if (wsn < 0) {
+					// Haven't reached the numbered workspaces
+					continue;
+				}
+				if (wsn < n && (!prev || wsn > prevn)) {
+					// The closest workspace before the current
+					prev = ws;
+					prevn = workspace_get_number(prev);
+				}
+			}
+		}
+	}
+
+	if (!prev) {
+		prev = other ? other : last;
+	}
+	return prev;
+}
+
+struct sway_workspace *workspace_next(struct sway_workspace *workspace) {
+	int n = workspace_get_number(workspace);
+	struct sway_workspace *next = NULL, *first = NULL, *other = NULL;
+	bool found = false;
+	if (n < 0) {
+		// Find the next named workspace
+		int othern = -1;
+		for (int i = 0; i < root->outputs->length; i++) {
+			struct sway_output *output = root->outputs->items[i];
+			for (int j = 0; j < output->workspaces->length; j++) {
+				struct sway_workspace *ws = output->workspaces->items[j];
+				int wsn = workspace_get_number(ws);
+				if (!first) {
+					// The first named workspace
+					first = ws;
+				}
+				if (!other || (wsn >= 0 && wsn < othern)) {
+					// The first (least) numbered workspace
+					other = ws;
+					othern = workspace_get_number(other);
+				}
+				if (ws == workspace) {
+					found = true;
+				} else if (wsn < 0 && found) {
+					// The first non-numbered workspace after the current
+					return ws;
+				}
+			}
+		}
+	} else {
+		// Find the next numbered workspace
+		int nextn = -1, firstn = -1;
+		for (int i = 0; i < root->outputs->length; i++) {
+			struct sway_output *output = root->outputs->items[i];
+			for (int j = 0; j < output->workspaces->length; j++) {
+				struct sway_workspace *ws = output->workspaces->items[j];
+				int wsn = workspace_get_number(ws);
+				if (!first || (wsn >= 0 && wsn < firstn)) {
+					// The first (or least numbered) workspace
+					first = ws;
+					firstn = workspace_get_number(first);
+				}
+				if (!other && wsn < 0) {
+					// The first non-numbered workspace
+					other = ws;
+				}
+				if (wsn < 0) {
+					// Checked all the numbered workspaces
+					break;
+				}
+				if (n < wsn && (!next || wsn < nextn)) {
+					// The first workspace numerically after the current
+					next = ws;
+					nextn = workspace_get_number(next);
+				}
+			}
+		}
+	}
+
+	if (!next) {
+		// If there is no next workspace from the same category, return the
+		// first from this category.
+		next = other ? other : first;
+	}
+	return next;
+}
+
 /**
  * Get the previous or next workspace on the specified output. Wraps around at
  * the end and beginning.  If next is false, the previous workspace is returned,
@@ -409,50 +551,16 @@ static struct sway_workspace *workspace_output_prev_next_impl(
 	return output->workspaces->items[new_index];
 }
 
-/**
- * Get the previous or next workspace. If the first/last workspace on an output
- * is active, proceed to the previous/next output's previous/next workspace.
- */
-static struct sway_workspace *workspace_prev_next_impl(
-		struct sway_workspace *workspace, int dir) {
-	struct sway_output *output = workspace->output;
-	int index = list_find(output->workspaces, workspace);
-	int new_index = index + dir;
-
-	if (new_index >= 0 && new_index < output->workspaces->length) {
-		return output->workspaces->items[new_index];
-	}
-
-	// Look on a different output
-	int output_index = list_find(root->outputs, output);
-	new_index = wrap(output_index + dir, root->outputs->length);
-	output = root->outputs->items[new_index];
-
-	if (dir == 1) {
-		return output->workspaces->items[0];
-	} else {
-		return output->workspaces->items[output->workspaces->length - 1];
-	}
-}
-
 struct sway_workspace *workspace_output_next(
 		struct sway_workspace *current, bool create) {
 	return workspace_output_prev_next_impl(current->output, 1, create);
 }
 
-struct sway_workspace *workspace_next(struct sway_workspace *current) {
-	return workspace_prev_next_impl(current, 1);
-}
-
 struct sway_workspace *workspace_output_prev(
 		struct sway_workspace *current, bool create) {
 	return workspace_output_prev_next_impl(current->output, -1, create);
 }
 
-struct sway_workspace *workspace_prev(struct sway_workspace *current) {
-	return workspace_prev_next_impl(current, -1);
-}
-
 bool workspace_switch(struct sway_workspace *workspace,
 		bool no_auto_back_and_forth) {
 	struct sway_seat *seat = input_manager_current_seat();
@@ -461,7 +569,7 @@ bool workspace_switch(struct sway_workspace *workspace,
 	if (focus && focus->type == N_WORKSPACE) {
 		active_ws = focus->sway_workspace;
 	} else if (focus && focus->type == N_CONTAINER) {
-		active_ws = focus->sway_container->workspace;
+		active_ws = focus->sway_container->pending.workspace;
 	}
 
 	if (!no_auto_back_and_forth && config->auto_back_and_forth && active_ws
@@ -498,7 +606,7 @@ bool workspace_is_empty(struct sway_workspace *ws) {
 	// Sticky views are not considered to be part of this workspace
 	for (int i = 0; i < ws->floating->length; ++i) {
 		struct sway_container *floater = ws->floating->items[i];
-		if (!floater->is_sticky) {
+		if (!container_is_sticky(floater)) {
 			return false;
 		}
 	}
@@ -627,16 +735,31 @@ struct sway_container *workspace_find_container(struct sway_workspace *ws,
 	return NULL;
 }
 
+static void set_workspace(struct sway_container *container, void *data) {
+	container->pending.workspace = container->pending.parent->pending.workspace;
+}
+
+static void workspace_attach_tiling(struct sway_workspace *ws,
+		struct sway_container *con) {
+	list_add(ws->tiling, con);
+	con->pending.workspace = ws;
+	container_for_each_child(con, set_workspace, NULL);
+	container_handle_fullscreen_reparent(con);
+	workspace_update_representation(ws);
+	node_set_dirty(&ws->node);
+	node_set_dirty(&con->node);
+}
+
 struct sway_container *workspace_wrap_children(struct sway_workspace *ws) {
 	struct sway_container *fs = ws->fullscreen;
 	struct sway_container *middle = container_create(NULL);
-	middle->layout = ws->layout;
+	middle->pending.layout = ws->layout;
 	while (ws->tiling->length) {
 		struct sway_container *child = ws->tiling->items[0];
 		container_detach(child);
 		container_add_child(middle, child);
 	}
-	workspace_add_tiling(ws, middle);
+	workspace_attach_tiling(ws, middle);
 	ws->fullscreen = fs;
 	return middle;
 }
@@ -648,9 +771,9 @@ void workspace_unwrap_children(struct sway_workspace *ws,
 		return;
 	}
 
-	ws->layout = wrap->layout;
-	while (wrap->children->length) {
-		struct sway_container *child = wrap->children->items[0];
+	ws->layout = wrap->pending.layout;
+	while (wrap->pending.children->length) {
+		struct sway_container *child = wrap->pending.children->items[0];
 		container_detach(child);
 		workspace_add_tiling(ws, child);
 	}
@@ -668,47 +791,41 @@ void workspace_detach(struct sway_workspace *workspace) {
 	node_set_dirty(&output->node);
 }
 
-static void set_workspace(struct sway_container *container, void *data) {
-	container->workspace = container->parent->workspace;
-}
-
-void workspace_add_tiling(struct sway_workspace *workspace,
+struct sway_container *workspace_add_tiling(struct sway_workspace *workspace,
 		struct sway_container *con) {
-	if (con->workspace) {
+	if (con->pending.workspace) {
 		container_detach(con);
 	}
+	if (config->default_layout != L_NONE) {
+		con = container_split(con, config->default_layout);
+	}
 	list_add(workspace->tiling, con);
-	con->workspace = workspace;
+	con->pending.workspace = workspace;
 	container_for_each_child(con, set_workspace, NULL);
 	container_handle_fullscreen_reparent(con);
 	workspace_update_representation(workspace);
 	node_set_dirty(&workspace->node);
 	node_set_dirty(&con->node);
+	return con;
 }
 
 void workspace_add_floating(struct sway_workspace *workspace,
 		struct sway_container *con) {
-	if (con->workspace) {
+	if (con->pending.workspace) {
 		container_detach(con);
 	}
 	list_add(workspace->floating, con);
-	con->workspace = workspace;
+	con->pending.workspace = workspace;
 	container_for_each_child(con, set_workspace, NULL);
 	container_handle_fullscreen_reparent(con);
 	node_set_dirty(&workspace->node);
 	node_set_dirty(&con->node);
 }
 
-void workspace_insert_tiling(struct sway_workspace *workspace,
+void workspace_insert_tiling_direct(struct sway_workspace *workspace,
 		struct sway_container *con, int index) {
-	if (con->workspace) {
-		container_detach(con);
-	}
-	if (workspace->layout == L_STACKED || workspace->layout == L_TABBED) {
-		con = container_split(con, workspace->layout);
-	}
 	list_insert(workspace->tiling, index, con);
-	con->workspace = workspace;
+	con->pending.workspace = workspace;
 	container_for_each_child(con, set_workspace, NULL);
 	container_handle_fullscreen_reparent(con);
 	workspace_update_representation(workspace);
@@ -716,6 +833,18 @@ void workspace_insert_tiling(struct sway_workspace *workspace,
 	node_set_dirty(&con->node);
 }
 
+struct sway_container *workspace_insert_tiling(struct sway_workspace *workspace,
+		struct sway_container *con, int index) {
+	if (con->pending.workspace) {
+		container_detach(con);
+	}
+	if (config->default_layout != L_NONE) {
+		con = container_split(con, config->default_layout);
+	}
+	workspace_insert_tiling_direct(workspace, con, index);
+	return con;
+}
+
 void workspace_add_gaps(struct sway_workspace *ws) {
 	if (config->smart_gaps) {
 		struct sway_seat *seat = input_manager_get_default_seat();
@@ -776,7 +905,7 @@ struct sway_container *workspace_split(struct sway_workspace *workspace,
 	enum sway_container_layout old_layout = workspace->layout;
 	struct sway_container *middle = workspace_wrap_children(workspace);
 	workspace->layout = layout;
-	middle->layout = old_layout;
+	middle->pending.layout = old_layout;
 
 	struct sway_seat *seat;
 	wl_list_for_each(seat, &server.input->seats, link) {
@@ -817,3 +946,23 @@ size_t workspace_num_tiling_views(struct sway_workspace *ws) {
 	workspace_for_each_container(ws, count_tiling_views, &count);
 	return count;
 }
+
+static void count_sticky_containers(struct sway_container *con, void *data) {
+	if (container_is_sticky(con)) {
+		size_t *count = data;
+		*count += 1;
+	}
+}
+
+size_t workspace_num_sticky_containers(struct sway_workspace *ws) {
+	size_t count = 0;
+	workspace_for_each_container(ws, count_sticky_containers, &count);
+	return count;
+}
+
+void workspace_squash(struct sway_workspace *workspace) {
+	for (int i = 0; i < workspace->tiling->length; i++) {
+		struct sway_container *child = workspace->tiling->items[i];
+		i += container_squash(child);
+	}
+}
